pg_reserved_words = [
    "ABORT",
    "ABS",
    "ABSOLUTE",
    "ACCESS",
    "ACTION",
    "ADA",
    "ADD",
    "ADMIN",
    "AFTER",
    "AGGREGATE",
    "ALIAS",
    "ALL",
    "ALLOCATE",
    "ALTER",
    "ANALYSE",
    "ANALYZE",
    "AND",
    "ANY",
    "ARE",
    "ARRAY",
    "AS",
    "ASC",
    "ASENSITIVE",
    "ASSERTION",
    "ASSIGNMENT",
    "ASYMMETRIC",
    "AT",
    "ATOMIC",
    "AUTHORIZATION",
    "AVG",
    "BACKWARD",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BIT",
    "BITVAR",
    "BIT_LENGTH",
    "BLOB",
    "BOOLEAN",
    "BOTH",
    "BREADTH",
    "BY",
    "CACHE",
    "CALL",
    "CALLED",
    "CARDINALITY",
    "CASCADE",
    "CASCADED",
    "CASE",
    "CAST",
    "CATALOG",
    "CATALOG_NAME",
    "CHAIN",
    "CHAR",
    "CHARACTER",
    "CHARACTERISTICS",
    "CHARACTER_LENGTH",
    "CHARACTER_SET_CATALOG",
    "CHARACTER_SET_NAME",
    "CHARACTER_SET_SCHEMA",
    "CHAR_LENGTH",
    "CHECK",
    "CHECKED",
    "CHECKPOINT",
    "CLASS",
    "CLASS_ORIGIN",
    "CLOB",
    "CLOSE",
    "CLUSTER",
    "COALESCE",
    "COBOL",
    "COLLATE",
    "COLLATION",
    "COLLATION_CATALOG",
    "COLLATION_NAME",
    "COLLATION_SCHEMA",
    "COLUMN",
    "COLUMN_NAME",
    "COMMAND_FUNCTION",
    "COMMAND_FUNCTION_CODE",
    "COMMENT",
    "COMMIT",
    "COMMITTED",
    "COMPLETION",
    "CONDITION_NUMBER",
    "CONNECT",
    "CONNECTION",
    "CONNECTION_NAME",
    "CONSTRAINT",
    "CONSTRAINTS",
    "CONSTRAINT_CATALOG",
    "CONSTRAINT_NAME",
    "CONSTRAINT_SCHEMA",
    "CONSTRUCTOR",
    "CONTAINS",
    "CONTINUE",
    "CONVERSION",
    "CONVERT",
    "COPY",
    "CORRESPONDING",
    "COUNT",
    "CREATE",
    "CREATEDB",
    "CREATEUSER",
    "CROSS",
    "CUBE",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_PATH",
    "CURRENT_ROLE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURSOR",
    "CURSOR_NAME",
    "CYCLE",
    "DATA",
    "DATABASE",
    "DATE",
    "DATETIME_INTERVAL_CODE",
    "DATETIME_INTERVAL_PRECISION",
    "DAY",
    "DEALLOCATE",
    "DEC",
    "DECIMAL",
    "DECLARE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DEFINED",
    "DEFINER",
    "DELETE",
    "DELIMITER",
    "DELIMITERS",
    "DEPTH",
    "DEREF",
    "DESC",
    "DESCRIBE",
    "DESCRIPTOR",
    "DESTROY",
    "DESTRUCTOR",
    "DETERMINISTIC",
    "DIAGNOSTICS",
    "DICTIONARY",
    "DISCONNECT",
    "DISPATCH",
    "DISTINCT",
    "DO",
    "DOMAIN",
    "DOUBLE",
    "DROP",
    "DYNAMIC",
    "DYNAMIC_FUNCTION",
    "DYNAMIC_FUNCTION_CODE",
    "EACH",
    "ELSE",
    "ENCODING",
    "ENCRYPTED",
    "END",
    "END",
    "EXEC",
    "EQUALS",
    "ESCAPE",
    "EVERY",
    "EXCEPT",
    "EXCEPTION",
    "EXCLUSIVE",
    "EXEC",
    "EXECUTE",
    "EXISTING",
    "EXISTS",
    "EXPLAIN",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FINAL",
    "FIRST",
    "FLOAT",
    "FOR",
    "FORCE",
    "FOREIGN",
    "FORTRAN",
    "FORWARD",
    "FOUND",
    "FREE",
    "FREEZE",
    "FROM",
    "FULL",
    "FUNCTION",
    "GENERAL",
    "GENERATED",
    "GET",
    "GLOBAL",
    "GO",
    "GOTO",
    "GRANT",
    "GRANTED",
    "GROUP",
    "GROUPING",
    "HANDLER",
    "HAVING",
    "HIERARCHY",
    "HOLD",
    "HOST",
    "HOUR",
    "IDENTITY",
    "IGNORE",
    "ILIKE",
    "IMMEDIATE",
    "IMMUTABLE",
    "IMPLEMENTATION",
    "IMPLICIT",
    "IN",
    "INCREMENT",
    "INDEX",
    "INDICATOR",
    "INFIX",
    "INHERITS",
    "INITIALIZE",
    "INITIALLY",
    "INNER",
    "INOUT",
    "INPUT",
    "INSENSITIVE",
    "INSERT",
    "INSTANCE",
    "INSTANTIABLE",
    "INSTEAD",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "INVOKER",
    "IS",
    "ISNULL",
    "ISOLATION",
    "ITERATE",
    "JOIN",
    "KEY",
    "KEY_MEMBER",
    "KEY_TYPE",
    "LANCOMPILER",
    "LANGUAGE",
    "LARGE",
    "LAST",
    "LATERAL",
    "LEADING",
    "LEFT",
    "LENGTH",
    "LESS",
    "LEVEL",
    "LIKE",
    "LIMIT",
    "LISTEN",
    "LOAD",
    "LOCAL",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCATION",
    "LOCATOR",
    "LOCK",
    "LOWER",
    "MAP",
    "MATCH",
    "MAX",
    "MAXVALUE",
    "MESSAGE_LENGTH",
    "MESSAGE_OCTET_LENGTH",
    "MESSAGE_TEXT",
    "METHOD",
    "MIN",
    "MINUTE",
    "MINVALUE",
    "MOD",
    "MODE",
    "MODIFIES",
    "MODIFY",
    "MODULE",
    "MONTH",
    "MORE",
    "MOVE",
    "MUMPS",
    "NAME",
    "NAMES",
    "NATIONAL",
    "NATURAL",
    "NCHAR",
    "NCLOB",
    "NEW",
    "NEXT",
    "NO",
    "NOCREATEDB",
    "NOCREATEUSER",
    "NONE",
    "NOT",
    "NOTHING",
    "NOTIFY",
    "NOTNULL",
    "NULL",
    "NULLABLE",
    "NULLIF",
    "NUMBER",
    "NUMERIC",
    "OBJECT",
    "OCTET_LENGTH",
    "OF",
    "OFF",
    "OFFSET",
    "OIDS",
    "OLD",
    "ON",
    "ONLY",
    "OPEN",
    "OPERATION",
    "OPERATOR",
    "OPTION",
    "OPTIONS",
    "OR",
    "ORDER",
    "ORDINALITY",
    "OUT",
    "OUTER",
    "OUTPUT",
    "OVERLAPS",
    "OVERLAY",
    "OVERRIDING",
    "OWNER",
    "PAD",
    "PARAMETER",
    "PARAMETERS",
    "PARAMETER_MODE",
    "PARAMETER_NAME",
    "PARAMETER_ORDINAL_POSITION",
    "PARAMETER_SPECIFIC_CATALOG",
    "PARAMETER_SPECIFIC_NAME",
    "PARAMETER_SPECIFIC_SCHEMA",
    "PARTIAL",
    "PASCAL",
    "PASSWORD",
    "PATH",
    "PENDANT",
    "PLACING",
    "PLI",
    "POSITION",
    "POSTFIX",
    "PRECISION",
    "PREFIX",
    "PREORDER",
    "PREPARE",
    "PRESERVE",
    "PRIMARY",
    "PRIOR",
    "PRIVILEGES",
    "PROCEDURAL",
    "PROCEDURE",
    "PUBLIC",
    "READ",
    "READS",
    "REAL",
    "RECHECK",
    "RECURSIVE",
    "REF",
    "REFERENCES",
    "REFERENCING",
    "REINDEX",
    "RELATIVE",
    "RENAME",
    "REPEATABLE",
    "REPLACE",
    "RESET",
    "RESTRICT",
    "RESULT",
    "RETURN",
    "RETURNED_LENGTH",
    "RETURNED_OCTET_LENGTH",
    "RETURNED_SQLSTATE",
    "RETURNS",
    "REVOKE",
    "RIGHT",
    "ROLE",
    "ROLLBACK",
    "ROLLUP",
    "ROUTINE",
    "ROUTINE_CATALOG",
    "ROUTINE_NAME",
    "ROUTINE_SCHEMA",
    "ROW",
    "ROWS",
    "ROW_COUNT",
    "RULE",
    "SAVEPOINT",
    "SCALE",
    "SCHEMA",
    "SCHEMA_NAME",
    "SCOPE",
    "SCROLL",
    "SEARCH",
    "SECOND",
    "SECTION",
    "SECURITY",
    "SELECT",
    "SELF",
    "SENSITIVE",
    "SEQUENCE",
    "SERIALIZABLE",
    "SERVER_NAME",
    "SESSION",
    "SESSION_USER",
    "SET",
    "SETOF",
    "SETS",
    "SHARE",
    "SHOW",
    "SIMILAR",
    "SIMPLE",
    "SIZE",
    "SMALLINT",
    "SOME",
    "SOURCE",
    "SPACE",
    "SPECIFIC",
    "SPECIFICTYPE",
    "SPECIFIC_NAME",
    "SQL",
    "SQLCODE",
    "SQLERROR",
    "SQLEXCEPTION",
    "SQLSTATE",
    "SQLWARNING",
    "STABLE",
    "START",
    "STATE",
    "STATEMENT",
    "STATIC",
    "STATISTICS",
    "STDIN",
    "STDOUT",
    "STORAGE",
    "STRICT",
    "STRUCTURE",
    "STYLE",
    "SUBCLASS_ORIGIN",
    "SUBLIST",
    "SUBSTRING",
    "SUM",
    "SYMMETRIC",
    "SYSID",
    "SYSTEM",
    "SYSTEM_USER",
    "TABLE",
    "TABLE_NAME",
    "TEMP",
    "TEMPLATE",
    "TEMPORARY",
    "TERMINATE",
    "THAN",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TIMEZONE_HOUR",
    "TIMEZONE_MINUTE",
    "TO",
    "TOAST",
    "TRAILING",
    "TRANSACTION",
    "TRANSACTIONS_COMMITTED",
    "TRANSACTIONS_ROLLED_BACK",
    "TRANSACTION_ACTIVE",
    "TRANSFORM",
    "TRANSFORMS",
    "TRANSLATE",
    "TRANSLATION",
    "TREAT",
    "TRIGGER",
    "TRIGGER_CATALOG",
    "TRIGGER_NAME",
    "TRIGGER_SCHEMA",
    "TRIM",
    "TRUE",
    "TRUNCATE",
    "TRUSTED",
    "TYPE",
    "UNCOMMITTED",
    "UNDER",
    "UNENCRYPTED",
    "UNION",
    "UNIQUE",
    "UNKNOWN",
    "UNLISTEN",
    "UNNAMED",
    "UNNEST",
    "UNTIL",
    "UPDATE",
    "UPPER",
    "USAGE",
    "USER",
    "USER_DEFINED_TYPE_CATALOG",
    "USER_DEFINED_TYPE_NAME",
    "USER_DEFINED_TYPE_SCHEMA",
    "USING",
    "VACUUM",
    "VALID",
    "VALIDATOR",
    "VALUE",
    "VALUES",
    "VARCHAR",
    "VARIABLE",
    "VARYING",
    "VERBOSE",
    "VERSION",
    "VIEW",
    "VOLATILE",
    "WHEN",
    "WHENEVER",
    "WHERE",
    "WITH",
    "WITHOUT",
    "WORK",
    "WRITE",
    "YEAR",
    "ZONE"
 ]


from typing import List, Optional

from fiddler.v2.constants import FiddlerTimestamp, FileType, UploadType
from fiddler.v2.schema.base import BaseDataSchema


class EventsIngest(BaseDataSchema):
    file_name: List[str]
    batch_id: Optional[str] = None
    events_schema: Optional[dict] = None
    id_field: Optional[str] = None
    is_update: Optional[bool] = None
    timestamp_field: Optional[str] = None
    timestamp_format: Optional[FiddlerTimestamp] = None
    group_by: Optional[str] = None
    file_type: Optional[FileType] = None
    upload_type = UploadType.EVENT

    class Config:
        use_enum_values = True


class EventIngest(BaseDataSchema):
    event: dict
    event_id: Optional[str] = None
    id_field: Optional[str] = None
    is_update: Optional[bool] = None
    event_timestamp: Optional[str] = None
    timestamp_format: Optional[FiddlerTimestamp] = None



import json
import os
import unittest
from io import BytesIO

from werkzeug.datastructures import FileStorage

from fiddler.file_processor.src.facade import upload_dataset

package_dir = os.path.dirname(os.path.abspath(__file__))


class FacadeTest(unittest.TestCase):
    def test_upload_dataset_with_valid_dataset_succeed(self):
        files = {}
        file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        file_name = 'credit_risk_dataset.csv'
        with open(file_path, 'rb') as fh:
            buf = BytesIO(fh.read())
            files = {file_name: FileStorage(buf, file_name)}

        schema_file_path = os.path.join(
            package_dir, 'resources/avro/simple_avro_mapping_schema.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        upload_dataset(files, 'LOCAL_DISK', 'avro', file_schema_str)

    def test_upload_dataset_with_invalid_dataset_throw_exception(self):
        files = {}
        file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        file_name = 'credit_risk_dataset.csv'
        with open(file_path, 'rb') as fh:
            buf = BytesIO(fh.read())
            files = {file_name: FileStorage(buf, file_name)}

        schema_file_path = os.path.join(
            package_dir, 'resources/avro/simple_avro_mapping_schema.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        self.assertRaises(
            ValueError, upload_dataset, files, 'LOCAL_DISK', 'avro1', file_schema_str
        )

    def test_upload_dataset_with_missing_avro_schema_dataset_throw_exception(self):
        files = {}
        file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        file_name = 'credit_risk_dataset.csv'
        with open(file_path, 'rb') as fh:
            buf = BytesIO(fh.read())
            files = {file_name: FileStorage(buf, file_name)}

        self.assertRaises(ValueError, upload_dataset, files, 'LOCAL_DISK', 'avro')


if __name__ == '__main__':
    unittest.main()



from typing import List, Optional

from fiddler.v2.constants import FileType, UploadType
from fiddler.v2.schema.base import BaseDataSchema
from fiddler.v2.schema.common import DatasetInfo


class Dataset(BaseDataSchema):

    id: int
    name: str
    version: str
    file_list: dict
    info: DatasetInfo
    organization_name: str
    project_name: str


class DatasetIngest(BaseDataSchema):
    name: str
    file_name: List[str]
    info: Optional[DatasetInfo] = None
    file_type: Optional[FileType] = None
    file_schema: Optional[dict] = None
    upload_type = UploadType.DATASET

    class Config:
        use_enum_values = True



import logging
from typing import Any, Dict, Optional

from fiddler.connection import Connection
from fiddler.project import Project

from ..core_objects import CURRENT_SCHEMA_VERSION, MalformedSchemaException, SegmentInfo
from ..utils.general_checks import type_enforce

LOG = logging.getLogger()


class Segment:
    def __init__(self, connection: Connection, project_id: str, model_id: str) -> None:
        self.connection = connection
        self.project_id = project_id
        self.model_id = model_id

    def list_segments(
        self,
        activated: Optional[bool] = None,
        schema_version: Optional[float] = None,
    ):
        """
        List all segments currently associated with a model. Can filter by activated / schema_version if desired.

        activated = [None/True/False], for all, activated-only, deactivate-only, respectively
        schema_verios = [None/0.1] for all, 0.1 respectively
        """
        res = None
        path = ['list_segments', self.connection.org_id, self.project_id, self.model_id]
        payload: Dict[str, Any] = {}
        payload['activated'] = activated
        payload['schema_version'] = schema_version
        res = self.connection.call(path, json_payload=payload)
        if res is None:
            LOG.warning('Could not get segments for this model.')
        return res

    def upload_segment(
        self,
        segment_id: str,
        segment_info: SegmentInfo,
    ):
        """Validates and uploads an existing SegmentInfo object to Fiddler."""

        model_info = None
        try:
            model_info = (
                Project(self.connection, self.project_id)
                .model(self.model_id)
                .get_info()
            )
        except Exception as e:
            LOG.exception(
                f'Did not find ModelInfo for project "{self.project_id}" and model "{self.model_id}".'
            )
            raise e

        try:
            assert self._validate_segment(segment_id, segment_info, model_info)
        except MalformedSchemaException as mse:
            LOG.exception(
                f'Could not validate SegmentInfo, does the object conform to schema version {CURRENT_SCHEMA_VERSION}?'
            )
            raise mse
        seg_info = segment_info.to_dict()

        res = None
        path = [
            'define_segment',
            self.connection.org_id,
            self.project_id,
            self.model_id,
            segment_id,
        ]
        res = self.connection.call(path, json_payload=seg_info)
        if res is None:
            LOG.warning('Could not create segment.')
        return res

    def activate_segment(self, segment_id: str):
        """Activate an existing segment in Fiddler"""
        segment_id = type_enforce('segment_id', segment_id, str)
        res = None
        path = [
            'activate_segment',
            self.connection.org_id,
            self.project_id,
            self.model_id,
            segment_id,
        ]
        res = self.connection.call(path)
        if res is None:
            LOG.warning('Could not activate segment.')
        return res

    def deactivate_segment(self, segment_id: str):
        """Deactivate an existing segment in Fiddler"""
        segment_id = type_enforce('segment_id', segment_id, str)
        res = None
        path = [
            'deactivate_segment',
            self.connection.org_id,
            self.project_id,
            self.model_id,
            segment_id,
        ]
        res = self.connection.call(path)
        if res is None:
            LOG.warning('Could not deactivate segment.')
        return res

    def delete_segment(self, segment_id: str):
        """Deletes a deactivated segments from Fiddler"""
        segment_id = type_enforce('segment_id', segment_id, str)
        res = None
        path = [
            'delete_segment',
            self.connection.org_id,
            self.project_id,
            self.model_id,
            segment_id,
        ]
        res = self.connection.call(path)
        if res is None:
            LOG.warning('Could not delete segment.')
        return res

    def get_segment_info(self, segment_id: str):
        """Get the SegmentInfo specified from Fiddler"""
        segment_id = type_enforce('segment_id', segment_id, str)
        res = None
        path = [
            'get_segment',
            self.connection.org_id,
            self.project_id,
            self.model_id,
            segment_id,
        ]
        res = self.connection.call(path)
        if res is None:
            LOG.exception('Could not get segment_info.')
            raise RuntimeError('res is unexpectedly None')
        if res['success']:
            return SegmentInfo.from_dict(res['segment_info'])
        else:
            return None

    def _validate_segment(self, segment_id, segment_info, model_info):
        """helper to validate a segment"""
        if segment_info.segment_id != segment_id:
            raise MalformedSchemaException(
                f'Specified segment_id={segment_id} != segment_info.segment_id={segment_info.segment_id}.'
            )
        return segment_info.validate(model_info)



from http import HTTPStatus
from typing import List

from pydantic import parse_obj_as

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.schema.project import Project
from fiddler.v2.utils.exceptions import handle_api_error_response
from fiddler.v2.utils.response_handler import (
    APIResponseHandler,
    PaginatedResponseHandler,
)

logger = logging.getLogger(__name__)


class ProjectMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def get_projects(self) -> List[Project]:
        """
        Get a list of all projects in the organization

        :returns: List of `Project` object
        """
        response = self.client.get(
            url='projects',
            params={'organization_name': self.organization_name},
        )
        _, items = PaginatedResponseHandler(response).get_pagination_details_and_items()
        return parse_obj_as(List[Project], items)

    @handle_api_error_response
    def delete_project(self, project_name: str) -> None:
        """
        Delete a project

        :params project_name: Name of the project to delete
        :returns: None
        """
        response = self.client.delete(
            url=f'projects/{self.organization_name}:{project_name}'
        )
        if response.status_code == HTTPStatus.OK:
            logger.info(f'{project_name} deleted successfully.')
        else:
            # @TODO: Handle non 200 status response
            logger.info('Delete unsuccessful')

    @handle_api_error_response
    def add_project(self, project_name: str) -> Project:
        """
        Add a new project.

        :param project_name: The unique identifier of the project on the
            Fiddler engine. Must be a short string without whitespace.

        :returns: Created `Project` object.
        """
        request_body = Project(
            name=project_name, organization_name=self.organization_name
        ).dict()
        response = self.client.post(
            url='projects',
            params={'organization_name': self.organization_name},
            data=request_body,
        )
        logger.info(f'{project_name} created successfully!')
        return Project.deserialize(APIResponseHandler(response))



import logging
import pandas as pd
import yaml

from fiddler.fiddler_api import FiddlerApi

LOG = logging.getLogger(__name__)

def execute_cmd(args):
    client = FiddlerApi(url=f'http://localhost:{args.port}', org_id=args.org)
    df = pd.read_csv('dataset.csv')
    if args.index:
        df = df.loc[[int(args.index)]]
    else:
        df = df.head()
    LOG.info('Input: ')
    LOG.info(df.head(5))
    result = client.run_model(args.project, args.model, df)
    LOG.info('Result: ')
    LOG.info(result)


def explain_cmd(args):
    client = FiddlerApi(url=f'http://localhost:{args.port}', org_id=args.org)
    df = pd.read_csv('dataset.csv')
    if args.index:
        df = df.loc[[int(args.index)]]
    else:
        df = df.head(1)
    LOG.info('Input: ')
    LOG.info(df.head(5))
    if args.explanations:
        explanations = args.explanations
    else:
        explanations = 'shap'
    result = client.run_explanation(
        args.project,
        args.model,
        df,
        dataset_id='titanic',
        explanations=explanations,
    )
    LOG.info('Output: ')
    LOG.info(yaml.dump(result))



import re
from typing import Union

import pandas as pd

from ..core_objects import BatchPublishType


def infer_data_source(source: Union[str, pd.DataFrame]):
    """
    Attempts to infer the type of object based on type
    """
    if isinstance(source, pd.DataFrame):
        source_type = BatchPublishType.DATAFRAME
    elif isinstance(source, str):
        if re.match(r'((s3-|s3\.)?(.*)\.amazonaws\.com|^s3://)', source):
            source_type = BatchPublishType.AWS_S3
        elif re.match(r'((gs-|gs\.)?(.*)\.cloud.google\.com|^gs://)', source):
            source_type = BatchPublishType.GCP_STORAGE
        else:
            source_type = BatchPublishType.LOCAL_DISK
    else:
        raise ValueError('Unable to infer BatchPublishType')

    return source_type



from collections import namedtuple
from http import HTTPStatus
from typing import List, NamedTuple

try:
    from simplejson import JSONDecodeError
except ImportError:
    from json import JSONDecodeError

from requests.exceptions import HTTPError

from fiddler.utils import logging

logger = logging.getLogger(__name__)


class ErrorResponseHandler:
    def __init__(self, http_error: HTTPError) -> None:
        self.http_error = http_error
        self.response = http_error.response
        self.ErrorResponse = namedtuple(
            'ErrorResponse', ['status_code', 'error_code', 'message', 'errors']
        )

    def get_error_details(self) -> NamedTuple:
        status_code = self.response.status_code
        try:
            error_details = self.response.json().get('error', {})
        except JSONDecodeError:
            raise FiddlerAPIException(
                status_code=self.response.status_code,
                error_code=self.response.status_code,
                message=f'Invalid response content-type. {self.response.status_code}:{self.response.content}',
                errors=[],
            )
        error_code = error_details.get('code')
        message = error_details.get('message')
        errors = error_details.get('errors')
        return self.ErrorResponse(status_code, error_code, message, errors)


class BaseException(Exception):
    pass


class FiddlerException(BaseException):
    '''
    Exception class to handle non API response exceptions
    '''


class FiddlerAPIException(BaseException):
    '''
    Exception class to specifically handle Fiddler's API error responses structure.
    This is a generic API response exception class
    '''

    # @TODO: Handle standard API error response.
    # How to surface error messages coming form the server. Server responds error messages in a list. Which error to surface?
    def __init__(
        self, status_code: int, error_code: int, message: str, errors: List[str]
    ) -> None:
        self.status_code = status_code
        self.error_code = error_code
        self.message = message
        self.errors = errors
        super().__init__(self.message)


class FiddlerAPIBadRequestException(FiddlerAPIException):
    pass


class FiddlerAPINotFoundException(FiddlerAPIException):
    pass


class FiddlerAPIConflictException(FiddlerAPIException):
    pass


class FiddlerAPIForbiddenException(FiddlerAPIException):
    pass


class FiddlerAPIInternalServerErrorException(FiddlerAPIException):
    pass


map_except_resp_code = {
    HTTPStatus.BAD_REQUEST: FiddlerAPIBadRequestException,  # 400
    HTTPStatus.FORBIDDEN: FiddlerAPIForbiddenException,  # 403
    HTTPStatus.NOT_FOUND: FiddlerAPINotFoundException,  # 404
    HTTPStatus.CONFLICT: FiddlerAPIConflictException,  # 409
    HTTPStatus.INTERNAL_SERVER_ERROR: FiddlerAPIInternalServerErrorException,  # 500
}


def handle_api_error_response(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPError as error:
            logger.error(f'Error occured at {func.__name__}')
            error_response = ErrorResponseHandler(error).get_error_details()
            # raise status_code specific exceptions else raise generic FiddlerAPIException
            exec_class = map_except_resp_code.get(
                error_response.status_code, FiddlerAPIException
            )
            raise exec_class(
                error_response.status_code,
                error_response.error_code,
                error_response.message,
                error_response.errors,
            ) from None
            # disabling automatic exception chaining
            # ref: https://docs.python.org/3/tutorial/errors.html#exception-chaining

    return wrapper



import pandas as pd


class TemplateKerasTF2Model:
    def __init__(self, model_dir, model_artifact_path, att_layer_names_mapping, output_col,
                 embedding_names=None, batch_input_shape_list=None, hybrid=False):
        """
        Template for Keras TF2 models for package.py creation. Each method can be overwrite if needed.

        :param model_dir: model directory pathlib path
        :param model_artifact_path: str. model artifact file name (if .h5 file format) or
               directory (if saved model format)
        :param att_layer_names_mapping: dict that map attributable layer names to input feature names.
               Values have to be list of feature name(s).
               For example: {'cat_var': ['country', 'gender'], 'num_var': ['price']}.
        :param output_col: List. Name(s) the output(s) of the model. This will need to match the output(s)
               defined in model info.
        :param embedding_names: List. Name(s) the embedding layer(s) of the model.
        :param batch_input_shape_list: List of tuples corresponding to the output shape of the embedding layer(s)
               of the model. For example if the model has two embeddings, the list should have two tuples.
        :param hybrid: Boolean value for hybrid data type (for example matrices). Default to False.
        """
        import tensorflow as tf
        from .keras_ig_helpers import ComputeGradientsKerasTF2

        self.output_col = output_col
        self.embedding_names = embedding_names
        self.hybrid = hybrid

        self.model = tf.keras.models.load_model(str(model_dir / model_artifact_path))

        self.grads_instance = ComputeGradientsKerasTF2(self.model, model_dir, model_artifact_path,
                                                       att_layer_names_mapping, embedding_names,
                                                       hybrid=self.hybrid)
        if (embedding_names is None) or (embedding_names == []):
            self.grad_model = self.model
        else:
            self.grad_model = self.grads_instance.define_model_grads(batch_input_shape_list)

    def transform_to_attributable_input(self, input_df):
        """
        This method is called by the platform and is responsible for transforming the input dataframe
        to the upstream-most representation of model inputs that belongs to a continuous vector-space.
        For models with embedding layers (esp. NLP models) the first attributable layer is downstream of that.

        :param input_df: pandas DataFrame
        :return: dictionary with keys attributable layer names and values array of corresponding values.
                 If there are embeddings, the attributable_input values should be the output of the embedding layer.
        """
        if (self.embedding_names is not None) or self.hybrid:
            input_df = self._transform_input(input_df)
        return self.grads_instance.transform_to_attributable_input(input_df)

    def get_ig_baseline(self, input_df):
        """
        This method is used to generate the baseline against which to compare the input.
        It accepts a pandas DataFrame object containing rows of raw feature vectors that
        need to be explained (in case e.g. the baseline must be sized according to the explain point).
        Must return a pandas DataFrame that can be consumed by the predict method described earlier.

        :param input_df: pandas DataFrame
        :return: pandas DataFrame
        """
        raise NotImplementedError('Please implement generate_baseline in package.py')

    def _transform_input(self, input_df):
        """
        Helper function that accepts a pandas DataFrame object containing rows of raw feature vectors.
        The output of this method can be any Python object. This function can also
        be used to deserialize complex data types stored in dataset columns (e.g. arrays, or images
        stored in a field in UTF-8 format).

        :param input_df: pandas DataFrame
        :return: transformed input data
        """
        raise NotImplementedError('Please implement _transform_input in package.py')

    def predict(self, input_df):
        """
        Basic predict wrapper.

        :param input_df: a pandas DataFrame of input features
        :return: a pandas DataFrame of predictions.
        """
        transformed_input = self._transform_input(input_df)
        pred = self.model.predict(transformed_input)
        return pd.DataFrame(pred, columns=self.output_col)

    def compute_gradients(self, attributable_input):
        """
        This method computes gradients of the model output wrt to the differentiable input.
        If there are embeddings, the attributable_input should be the output of the embedding
        layer. In the backend, this method receives the output of the transform_to_attributable_input()
        method.

        :param attributable_input: dictionary with keys the attributable input layer names and values
               the differential input values associated.
        :return: a list of dictionaries, where each entry of the list is the attribution
                 for an output.
                 In case of single output model, this is a list with a single entry.
                 For the dictionary, the keys are the name of the input layers and the values are the attributions.
        """
        gradients_by_output = self.grads_instance.compute_gradients(attributable_input, self.grad_model)
        return gradients_by_output

    def project_attributions(self, input_df, attributions):
        """
        This method is used to map the attributions that were computed back to the input data.
        If you have a text input, you can use IGTextAttributionsTF2Keras.get_project_attribution().
        If you have tabular inputs, you can use IGTabularAttributions.get_project_attribution().
        If you have hybrid inputs, please contact Fiddler for help.

        :param input_df: pandas DataFrame of input features
        :param attributions: a list of dictionaries, where each entry of the list is the IG attribution for an output.
        :return: dictionary of final attributions for Fiddler to process. Keys are model outputs and values
                 are constructed from GEM (Generalized Explanation Markup) objects.
        """
        raise NotImplementedError('Please implement project_attributions in package.py. '
                                  'If you have a text input, you can use '
                                  'IGTextAttributionsTF2Keras.get_project_attribution(). '
                                  'If you have tabular inputs, you can use '
                                  'IGTabularAttributions.get_project_attribution(). '
                                  'If you have hybrid inputs, please contact Fiddler for help.')


class TemplateTreeShap:
    def __init__(self, model, output_cols, baseline_pred_dic, transformed_names_mapping=None):
        self.model = model
        self.output_cols = output_cols
        self.baseline_pred_dic = baseline_pred_dic
        self.transformed_names_mapping = transformed_names_mapping
        self.tree_shap_explanations = 'Tree Shap'

    def predict(self, input_df):
        """
        Basic predict wrapper.

        :param input_df: a pandas DataFrame of input features
        :return: a pandas DataFrame of predictions.
        """
        NotImplementedError('Please implement predict in package.py')

    def transform_input(self, input_df):
        """
        Basic wrapper to pre-process data.

        :param input_df: a pandas DataFrame of input features
        :return: a pandas DataFrame of predictions.
        """
        return input_df

    def explain_custom(self, explanation_name, input_df):
        """
        Wrapper to define the custom explanation method, in this example Tree SHAP.

        :param explanation_name: parameter used in Fiddler BE code to get explanation for a given explanation method.
               This allows users to define multiple custom explanations methods.
        :param input_df: a pandas DataFrame of input features
        :return: explanations_by_output and extras_by_output: two dictionaries with keys the output columns and values
                 the corresponding explanations and extra information necessary
        """
        from .project_attributions_helpers import TreeShapAttributions

        if explanation_name == self.tree_shap_explanations:
            predictions = self.predict(input_df).iloc[0].to_dict()
            transformed_df = self.transform_input(input_df)
            tree_shap_init = TreeShapAttributions(input_df, transformed_df, self.output_cols, self.model)
            if (self.transformed_names_mapping is not None) and (not isinstance(self.transformed_names_mapping, dict)):
                try:
                    self.transformed_names_mapping = self.transformed_names_mapping.get_mapping_feature_names(transformed_df.columns)
                except AttributeError:
                    raise ValueError('transformed_names_mapping is not a mapping dictionary')
            explanations_by_output, extras_by_output = tree_shap_init.get_project_attribution(predictions,
                                                                                              self.baseline_pred_dic,
                                                                                              self.transformed_names_mapping)
            return explanations_by_output, extras_by_output
        else:
            raise NotImplementedError(f'Please implement {explanation_name} in package.py with explain_custom method')



import enum


class FiddlerPublishSchema:
    STATIC = '__static'
    DYNAMIC = '__dynamic'
    ITERATOR = '__iterator'
    UNASSIGNED = '__unassigned'

    VALID_SCHEMA_KEYWORDS = [STATIC, DYNAMIC, ITERATOR]

    ORG = '__org'
    MODEL = '__model'
    PROJECT = '__project'
    TIMESTAMP = '__timestamp'
    DEFAULT_TIMESTAMP = '__default_timestamp'
    TIMESTAMP_FORMAT = '__timestamp_format'
    EVENT_ID = '__event_id'
    IS_UPDATE_EVENT = '__is_update_event'
    STATUS = '__status'
    LATENCY = '__latency'
    ITERATOR_KEY = '__iterator_key'

    CURRENT_TIME = 'CURRENT_TIME'


@enum.unique
class DataUploadType(enum.Enum):
    """Supported Data Upload for the Fiddler engine."""

    DATAFRAME = 0
    LOCAL_DISK = 1
    AWS_S3 = 2
    GCP_STORAGE = 3


class ExtractorDataSource:
    UNKNOWN = -1
    JSON = 0
    LOCAL_DISK = DataUploadType.LOCAL_DISK.value
    AWS_S3 = DataUploadType.AWS_S3.value
    GCP_STORAGE = DataUploadType.GCP_STORAGE.value


class ExtractorDataType:
    UNKNOWN = -1
    BATCHED = 0
    STREAMED = 1


class ExtractorFileType:
    UNKNOWN = -1
    CSV = 0
    PARQUET = 1
    PKL = 2
    CSV_GZ = 3
    AVRO = 4


class TransformType:
    UNKNOWN = -1
    JSON = 0
    DF = 1
    PARQUET = 2
    AVRO = 3


# Since `None` is a valid return type, this value is used to indicate a strictly missing value.
#  e.g. a dictionary with no key would return `SENTINEL_ABSENT`
SENTINEL_ABSENT = ()


SUPPORTABLE_FILE_EXTENSIONS = ['.csv', '.avro']
FILE_EXTENSIONS_TO_CODE = {
    '.csv': ExtractorFileType.CSV,
    '.csv.gz': ExtractorFileType.CSV_GZ,
    '.parquet': ExtractorFileType.PARQUET,
    '.pq': ExtractorFileType.PARQUET,
    '.pkl': ExtractorFileType.PKL,
    '.avro': ExtractorFileType.AVRO,
}
DTYPES_CSV = 'dtypes.csv'
TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

CHUNK_SIZE = 1000
DOT = '.'

# CSV configurations
CSV_EXTENSION = 'csv'

# Parquet configuration
PARQUET_EXTENSION = 'parquet'
PARQUET_COMPRESSION = 'snappy'
PARQUET_ENGINE = 'pyarrow'
PARQUET_ROW_GROUP_SIZE = 10_000



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import yaml

from fiddler.core_objects import BatchPublishType, Column, DatasetInfo, FiddlerTimestamp
from fiddler.utils import logging
from fiddler.v2.api.api import Client
from fiddler.v2.constants import FiddlerTimestamp as V2FiddlerTimestamp
from fiddler.v2.constants import FileType
from fiddler.v2.schema.common import Column as V2Column
from fiddler.v2.schema.common import DatasetInfo as V2DatasetInfo
from fiddler.v2.schema.dataset import Dataset

DATASET_MAX_ROWS = 50_000


logger = logging.getLogger(__name__)


class V1V2Compat:
    def __init__(self, client_v2: Client):
        self.client_v2 = client_v2

    # Projects
    def get_projects(self, get_project_details: bool = False) -> List[str]:
        """List the ids of all projects in the organization.

        :param get_project_details: Unused argument to maintain compatibility
        :returns: List of strings containing the ids of each project.
        """
        projects = self.client_v2.get_projects()
        return [p.name for p in projects]

    def add_project(self, project_id: str) -> Dict[str, str]:
        """Create a new project.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine. Must be a short string without whitespace.

        :returns: Server response for creation action.
        """
        project = self.client_v2.add_project(project_name=project_id)
        return {'project_name': project.name}

    def delete_project(self, project_id: str) -> None:
        """Permanently delete a project.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.

        :returns: None
        """
        self.client_v2.delete_project(project_name=project_id)

    # Datasets
    def get_datasets(self, project_id: str) -> List[str]:
        """List the ids of all datasets in the organization.

        :param project_id: The unique identifier of the project on Fiddler
        :returns: List of strings containing the ids of each dataset.
        """
        datasets = self.client_v2.get_datasets(project_name=project_id)
        return [d.name for d in datasets]

    def get_dataset_artifact(
        self,
        project_id: str,
        dataset_id: str,
        max_rows: int = 1_000,
        splits: Optional[List[str]] = None,
        sampling=False,
        dataset_info: Optional[DatasetInfo] = None,
        include_fiddler_id=False,
    ) -> Dict[str, pd.DataFrame]:
        raise NotImplementedError('This method is currently not implemented')

    def get_dataset(self, project_id: str, dataset_id: str) -> Dataset:
        dataset = self.client_v2.get_dataset(
            project_name=project_id, dataset_name=dataset_id
        )
        return dataset  # @todo: Make the response compatible with dataset info

    def get_dataset_info(self, project_id: str, dataset_id: str) -> DatasetInfo:
        """Get DatasetInfo for a dataset.

        :param project_id: The unique identifier of the project on Fiddler
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: A fiddler.DatasetInfo object describing the dataset.
        """
        dataset = self.get_dataset(project_id, dataset_id)
        dataset_info = DatasetInfo(
            display_name=dataset.name,
            columns=[
                Column.from_dict(col.dict(by_alias=True))
                for col in dataset.info.columns
            ],
            dataset_id=dataset_id,
        )
        return dataset_info

    def delete_dataset(self, project_id: str, dataset_id: str) -> None:
        """Permanently delete a dataset.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: None
        """
        # @todo: FDL-5366
        # check that on the server side
        # 1. Delete the table in CH
        # 2. Delete the files from blob store
        # 3. Delete the entry from postgres table
        self.client_v2.delete_dataset(project_name=project_id, dataset_name=dataset_id)

    # Uploading Dataset
    def upload_dataset_dataframe(
        self,
        project_id: str,
        dataset: Dict[str, pd.DataFrame],
        dataset_id: str,
        info: Optional[DatasetInfo] = None,
        size_check_enabled: bool = True,
    ) -> Dict[str, str]:
        """Uploads a representative dataset to the Fiddler engine.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param dataset: A dictionary mapping name -> DataFrame
            containing data to be uploaded to the Fiddler engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine. Must be a short string without whitespace.
        :param info: A DatasetInfo object specifying all the details of the
            dataset.
        :param size_check_enabled: Unused argument to maintain compatibility

        :returns: The server response for the upload.
        """
        v2_info = None
        if info:
            v2_info = V2DatasetInfo(
                columns=[V2Column.from_dict(col.to_dict()) for col in info.columns]
            )

        return self.client_v2.upload_dataset_dataframe(
            project_name=project_id,
            dataset_name=dataset_id,
            datasets=dataset,
            info=v2_info,
            is_sync=True,
        )

    def upload_dataset(
        self,
        project_id: str,
        dataset_id: str,
        file_path: str,
        file_type: str = 'csv',
        file_schema=Dict[str, Any],
        info: Optional[DatasetInfo] = None,
        size_check_enabled: bool = False,
    ) -> Dict[str, str]:
        """Uploads a representative dataset to the Fiddler engine from directory.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine. Must be a short string without whitespace.
        :param file_path: str pointing to the dataset to be uploaded
        :param file_type: str representing the file type of the uploading dataset. Supported type 'csv'
        :param file_schema: <TBD>
        :param info: A DatasetInfo object specifying all the details of the
            dataset.
        :param size_check_enabled: Unused argument to maintain compatibility

        :returns: The server response for the upload.
        """
        file_name = file_path.split('/')[-1]
        files = {file_name: Path(file_path)}
        # @todo as we only support csv on server side for now, throw an error saying the same.
        v2_info = None
        if info:
            v2_info = V2DatasetInfo(
                columns=[V2Column.from_dict(col.to_dict()) for col in info.columns]
            )
        if file_type == 'csv':
            return self.client_v2.upload_dataset(
                project_name=project_id,
                dataset_name=dataset_id,
                files=files,
                info=v2_info,
                file_type=FileType.CSV,
                file_schema=file_schema,
            )
        else:
            raise NotImplementedError('Only csv is supported in current implementation')

    def upload_dataset_dir(
        self,
        project_id: str,
        dataset_id: str,
        dataset_dir: Path,
        file_type: str = 'csv',
        file_schema=None,
        size_check_enabled: bool = False,
    ) -> Dict[str, str]:
        # Input checks
        if file_type != 'csv':
            raise NotImplementedError('Only CSV filetype is supported')

        if not dataset_dir.is_dir():
            raise ValueError(f'{dataset_dir} is not a directory')

        dataset_yaml = dataset_dir / f'{dataset_id}.yaml'
        if not dataset_yaml.is_file():
            raise ValueError(f'Dataset YAML file not found: {dataset_yaml}')

        with dataset_yaml.open() as f:
            dataset_info = DatasetInfo.from_dict(yaml.safe_load(f))

        # Convert files into dataframes
        files = dataset_dir.glob(f'*.{FileType.CSV}')
        csv_files = [x for x in files if x.is_file()]

        dataset = {}
        csv_paths = []
        for file in csv_files:
            csv_name = str(file).split('/')[-1]
            csv_paths.append(csv_name)
            name = csv_name[:-4]

            # @TODO Change the flow so that we can read the CSV in chunks
            dataset[name] = pd.read_csv(file, dtype=dataset_info.get_pandas_dtypes())

        # size check
        size_exceeds = False
        for name, df in dataset.items():
            if df.shape[0] > DATASET_MAX_ROWS:
                size_exceeds = True
        if size_exceeds:
            raise RuntimeError(
                f'Dataset upload aborted as size exceeds {DATASET_MAX_ROWS}.'
            )

        return self.upload_dataset_dataframe(
            project_id=project_id,
            dataset_id=dataset_id,
            dataset=dataset,
            info=dataset_info,
            size_check_enabled=size_check_enabled,
        )

    def upload_dataset_from_dir(
        self,
        project_id: str,
        dataset_id: str,
        dataset_dir: Path,
        file_type: str = 'csv',
        file_schema=None,
        size_check_enabled: bool = False,
    ) -> Dict[str, str]:
        """Uploads a representative dataset to the Fiddler engine from directory.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine. Must be a short string without whitespace.
        :param dataset_dir: pathlib.Path pointing to the dir that contains the dataset
            and dataset info yaml to be uploaded.
        :param file_type: File type of the dataset being uploaded. Supported type: 'csv'
        :param file_schema: <TBD>
        :param size_check_enabled: Unused argument to maintain compatibility

        :returns: The server response for the upload.
        """
        if file_type != 'csv':
            raise NotImplementedError('Only CSV filetype is supported')

        info_yaml = dataset_dir / f'{dataset_id}.yaml'

        if not info_yaml.exists():
            raise ValueError(f'DatasetInfo yaml ({info_yaml}) not found.')

        with open(info_yaml) as f:
            dataset_info = DatasetInfo.from_dict(yaml.safe_load(f))

        v2_info = V2DatasetInfo(
            columns=[V2Column.from_dict(col.to_dict()) for col in dataset_info.columns]
        )

        return self.client_v2.upload_dataset_from_dir(
            project_id,
            dataset_id,
            dataset_dir,
            v2_info,
            FileType.CSV,
            file_schema,
            is_sync=True,
        )
        
    def publish_events_batch_schema(self, events_path: Path, schema: Dict[str, Any]) -> None:
        return self.client_v2.publish_events_batch_schema(events_path, schema)

    def publish_events_batch(  # noqa
        self,
        project_id: str,
        model_id: str,
        batch_source: Union[pd.DataFrame, str],
        id_field: Optional[str] = None,
        update_event: Optional[bool] = False,
        timestamp_field: Optional[str] = None,
        timestamp_format: FiddlerTimestamp = FiddlerTimestamp.INFER,
        data_source: Optional[BatchPublishType] = None,
        casting_type: Optional[bool] = False,
        credentials: Optional[dict] = None,
        group_by: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Publishes a batch events object to Fiddler Service.
        :param project_id:    The project to which the model whose events are being published belongs.
        :param model_id:      The model whose events are being published.
        :param batch_source:  Batch object to be published. Can be one of: Pandas DataFrame, CSV file
        :param id_field:  Column to extract id value from.
        :param update_event: Bool indicating if the events are updates to previously published rows
        :param timestamp_field:     Column to extract timestamp value from.
                              Timestamp must match the specified format in `timestamp_format`.
        :param timestamp_format:   Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param data_source:   Source of batch object. In case of failed inference, can be one of:
                                - BatchPublishType.DATAFRAME
                                - BatchPublishType.LOCAL_DISK
                                - BatchPublishType.AWS_S3
                                - BatchPublishType.GCP_STORAGE
        :param casting_type: Unused argument to maintain compatibility
        :param credentials: Unused argument to maintain compatibility
        :param group_by: Column to group events together for Model Performance metrics. For example,
                         in ranking models that column should be query_id or session_id, used to
                         compute NDCG and MAP. Be aware that the batch_source file/dataset provided should have
                         events belonging to the SAME query_id/session_id TOGETHER and cannot be mixed
                         in the file. For example, having a file with rows belonging to query_id 31,31,31,2,2,31,31,31
                         would not work. Please sort the file by group_by group first to have rows with
                         the following order: query_id 31,31,31,31,31,31,2,2.
        """
        v2_timestamp_format = V2FiddlerTimestamp(timestamp_format.value)
        if type(batch_source) == pd.DataFrame and (
            data_source is None or BatchPublishType.DATAFRAME == data_source
        ):
            return self.client_v2.publish_events_batch_dataframe(
                project_name=project_id,
                model_name=model_id,
                events_df=batch_source,
                id_field=id_field,
                is_update=update_event,
                timestamp_field=timestamp_field,
                timestamp_format=v2_timestamp_format,
                group_by=group_by,
                is_sync=True,
            )
        elif type(batch_source) == str and (
            data_source is None or BatchPublishType.LOCAL_DISK == data_source
        ):
            return self.client_v2.publish_events_batch(
                project_name=project_id,
                model_name=model_id,
                events_path=Path(batch_source),
                id_field=id_field,
                is_update=update_event,
                timestamp_field=timestamp_field,
                timestamp_format=v2_timestamp_format,
                group_by=group_by,
                file_type=FileType.CSV,
                is_sync=True,
            )
        else:
            raise NotImplementedError(
                'Batch source other than dataframe and csv is not implemented now'
            )

    def publish_event(
        self,
        project_id: str,
        model_id: str,
        event: dict,
        event_id: Optional[str] = None,
        update_event: Optional[bool] = None,
        event_timestamp: Optional[int] = None,
        timestamp_format: FiddlerTimestamp = FiddlerTimestamp.INFER,
        casting_type: Optional[bool] = False,
        dry_run: Optional[bool] = False,
    ) -> str:
        """
        Publishes an event to Fiddler Service.
        :param project_id: The project to which the model whose events are being published belongs
        :param model_id: The model whose events are being published
        :param dict event: Dictionary of event details, such as features and predictions.
        :param event_id: Unique str event id for the event
        :param update_event: Bool indicating if the event is an update to a previously published row
        :param event_timestamp: The UTC timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param timestamp_format:   Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param casting_type: Unused argument to maintain compatibility
        :param dry_run: Unused argument to maintain compatibility

        """
        v2_timestamp_format = V2FiddlerTimestamp(timestamp_format.value)
        return self.client_v2.publish_event(
            project_name=project_id,
            model_name=model_id,
            event=event,
            event_id=event_id,
            is_update=update_event,
            event_timestamp=event_timestamp,
            timestamp_format=v2_timestamp_format,
        )



import unittest
from io import BytesIO

import importlib_resources
from werkzeug.datastructures import FileStorage

from fiddler.file_processor.src.constants import ExtractorFileType
from fiddler.file_processor.src.extractor import FileExtractor


class LocalFileExtractorTest(unittest.TestCase):
    def test_local_disk_extractor_with_single_file_should_succeed(self):
        # GIVEN
        files = {}
        filepath_context_manager = importlib_resources.path(
            'fiddler.file_processor.tests.resources.csvs', 'credit_risk_dataset.csv'
        )
        file_name = 'credit_risk_dataset.csv'
        with filepath_context_manager as fh:
            buf = BytesIO(fh.read_bytes())

            files = {file_name: FileStorage(buf, file_name)}

        # WHEN
        extractor = FileExtractor.get_extractor('LOCAL_DISK')
        file_list = extractor.extract(files)

        # THEN
        for temp_dir, file_path, file_extension in file_list:
            assert file_extension == ExtractorFileType.CSV
            assert file_path.split('/')[-1] == file_name

        assert len(file_list) == 1

    def test_local_disk_extractor_with_multiple_files_should_succeed(self):
        # GIVEN
        files = {}
        filepath_context_manager = importlib_resources.path(
            'fiddler.file_processor.tests.resources.csvs', 'credit_risk_dataset.csv'
        )
        second_filepath_context_manager = importlib_resources.path(
            'fiddler.file_processor.tests.resources.csvs', 'credit_risk_dataset_1.csv'
        )
        first_file_name = 'credit_risk_dataset.csv'
        second_file_name = 'credit_risk_dataset_1.csv'
        with filepath_context_manager as fh:
            first_file_buffer = BytesIO(fh.read_bytes())

        with second_filepath_context_manager as fh:
            second_file_buffer = BytesIO(fh.read_bytes())

            files = {
                first_file_name: FileStorage(first_file_buffer, first_file_name),
                second_file_name: FileStorage(second_file_buffer, second_file_name),
            }

        # WHEN
        extractor = FileExtractor.get_extractor('LOCAL_DISK')
        file_list = extractor.extract(files)

        # THEN
        for temp_dir, file_path, file_extension in file_list:
            assert file_extension == ExtractorFileType.CSV
            assert file_path.split('/')[-1] in [first_file_name, second_file_name]

        assert len(file_list) == 2


if __name__ == '__main__':
    unittest.main()



from typing import Optional

from fiddler.v2.schema.base import BaseDataSchema


class Model(BaseDataSchema):
    id: Optional[int] = None
    name: str
    project_name: str
    organization_name: str
    # @TODO: OAS says info is optional. Check it.
    info: Optional[dict] = None
    file_list: Optional[dict] = None
    model_type: Optional[str] = None
    framework: Optional[str] = None
    requirements: Optional[str] = None
    # @TODO: Return deployment obj instead of deployment_id
    model_deployment_id: Optional[int] = None



import copy
import logging
from typing import Dict, List

import pandas as pd

from .. import constants
from ..core_objects import Column, DatasetInfo, DataType, ModelInfo
from .pandas import is_datetime, try_series_retype

LOG = logging.getLogger(__name__)
pd.set_option('mode.chained_assignment', None)


def df_from_json_rows(
    dataset_rows_json: List[dict],
    dataset_info: DatasetInfo,
    include_fiddler_id: bool = False,
) -> pd.DataFrame:
    """Converts deserialized JSON into a pandas DataFrame according to a
        DatasetInfo object.

    If `include_fiddler_id` is true, we assume there is an extra column at
    the zeroth position containing the fiddler ID.
    """
    column_names = dataset_info.get_column_names()
    if include_fiddler_id:
        column_names.insert(0, constants.DATASET_FIDDLER_ID)
    if include_fiddler_id:
        dataset_info = copy.deepcopy(dataset_info)
        dataset_info.columns.insert(
            0, Column(constants.DATASET_FIDDLER_ID, DataType.STRING)
        )
    df = pd.DataFrame(dataset_rows_json, columns=dataset_info.get_column_names())
    for column_name in df:
        dtype = dataset_info[column_name].get_pandas_dtype()
        df[column_name] = try_series_retype(df[column_name], dtype)
    return df


def retype_df_for_model(df: pd.DataFrame, model_info: ModelInfo) -> pd.DataFrame:
    all_columns = (
        model_info.inputs
        if model_info.targets is None
        else model_info.inputs + model_info.targets
    )
    for column in all_columns:
        if column.name in df:
            df[column.name] = try_series_retype(
                df[column.name], column.get_pandas_dtype()
            )
    return df


def cast_input_data(data, model_info, fast_fail=True):
    """
    :param data: dictionary or pandas dataframe. Data we will cast with respect to the model info types
    :param model_info: info for the model from ModelInfo.
    :param fast_fail: Bool determining if violations throw error, or are ignored

    :return:
    """
    columns = model_info.inputs + model_info.targets + model_info.outputs
    col_mapping = {col.name: col.data_type.value for col in columns}
    possible_values_mapping = {col.name: col.possible_values for col in columns}
    is_dic = False
    if isinstance(data, dict):
        # publish_event is a dictionary
        data = pd.DataFrame.from_dict([data])
        is_dic = True
    for col in data.columns:
        if col in col_mapping.keys():
            col_type = col_mapping[col]
            if col_type == DataType.INTEGER.value:
                cast_type = int
            elif col_type == DataType.FLOAT.value:
                cast_type = float
            elif col_type == DataType.BOOLEAN.value:
                cast_type = bool
            else:
                # for category and string
                cast_type = str
            try:
                data.loc[:, col] = data.loc[:, col].astype(cast_type)
            except ValueError:
                if fast_fail:
                    raise TypeError(
                        f'Type casting failed for variable {col}. '
                        f'Model requires values to be {cast_type}.'
                    )
            if fast_fail:
                if col_type in [DataType.CATEGORY.value, DataType.BOOLEAN.value]:
                    if not data[col].values[0] in possible_values_mapping[col]:
                        raise ValueError(
                            f'Type casting failed for variable {col}. '
                            f'Model requires values to be in {possible_values_mapping[col]}.'
                            f' But found "{data[col].values[0]}" of type {type(data[col].values[0])}'
                        )
    if is_dic:
        data = data.to_dict(orient='records')[0]
    return data


class ColorLogger:

    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    WHITE = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    def _log(self, message, color):
        print(f'{color}{message}{self.WHITE}')

    def info(self, message):
        self._log(message, self.BLUE)

    def success(self, message):
        self._log(message, self.GREEN)

    def error(self, message):
        self._log(message, self.RED)

    def warn(self, message):
        self._log(message, self.YELLOW)



from typing import Dict, List, Tuple

try:
    from simplejson import JSONDecodeError
except ImportError:
    from json import JSONDecodeError

import requests

from fiddler.utils import logging
from fiddler.v2.utils.exceptions import FiddlerAPIException

logger = logging.getLogger(__name__)


class BaseResponseHandler:
    def __init__(self, response: requests.Response) -> None:
        self.response = response
        logger.debug(self.response.content)

    def get_data(self) -> Dict[str, dict]:
        try:
            dict_response = self.response.json().get('data')
        except JSONDecodeError:
            raise FiddlerAPIException(
                status_code=self.response.status_code,
                error_code=self.response.status_code,
                message=f'Invalid response content-type. {self.response.status_code}:{self.response.content}',
                errors=[],
            )
        dict_response.pop('created_at', None)
        dict_response.pop('updated_at', None)
        return dict_response

    def get_status_code(self) -> int:
        return self.response.status_code


class PaginatedResponseHandler(BaseResponseHandler):
    '''
    Handle fiddler OAS's standard Paginated response
    '''

    def get_pagination_details_and_items(self) -> Tuple[dict, List[dict]]:
        data = self.get_data()
        items = data.pop('items')
        return data, items


class APIResponseHandler(BaseResponseHandler):
    '''
    Handle fiddler OAS's standard API Response
    '''


class JobResponseHandler(BaseResponseHandler):
    def __init__(self, response: requests.Response) -> None:
        super().__init__(response)
        data = self.get_data()
        self.uuid = data.get('uuid')
        self.name = data.get('name')
        self.status = data.get('status')
        self.progress = data.get('progress')
        self.error_message = data.get('error_message')
        self.extras = data.get('extras')



import json
from copy import deepcopy
from typing import Dict, Optional
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter

APP_JSON_CONTENT_TYPE = 'application/json'


class RequestClient:
    def __init__(self, base_url: str, headers: Dict[str, str]) -> None:
        self.base_url = base_url
        self.headers = headers
        self.headers.update({'Content-Type': APP_JSON_CONTENT_TYPE})
        self.session = requests.Session()
        adapter = HTTPAdapter(
            pool_connections=25,
            pool_maxsize=25,
        )
        self.session.mount(self.base_url, adapter)

    def call(
        self,
        *,
        method: str,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ) -> requests.Response:
        """
        Make request to server
        """
        full_url = urljoin(self.base_url + '/', url)

        request_headers = self.headers
        # override/update headers coming from the calling method
        if headers:
            request_headers = deepcopy(self.headers)
            request_headers.update(headers)

        content_type = request_headers.get('Content-Type')
        if content_type == APP_JSON_CONTENT_TYPE:
            data = json.dumps(data)
        response = self.session.request(
            method,
            full_url,
            params=params,
            data=data,
            headers=request_headers,
            timeout=timeout,
            **kwargs,
        )
        response.raise_for_status()
        return response

    def get(
        self,
        *,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        return self.call(
            method='GET',
            url=url,
            params=params,
            headers=headers,
            timeout=timeout,
            **kwargs,
        )

    def delete(
        self,
        *,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        return self.call(
            method='DELETE',
            url=url,
            params=params,
            headers=headers,
            timeout=timeout,
            **kwargs,
        )

    def post(
        self,
        *,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        return self.call(
            method='POST',
            url=url,
            params=params,
            headers=headers,
            timeout=timeout,
            data=data,
            **kwargs,
        )

    def put(
        self,
        *,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        return self.call(
            method='PUT',
            url=url,
            params=params,
            headers=headers,
            timeout=timeout,
            data=data,
            **kwargs,
        )

    def patch(
        self,
        *,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        return self.call(
            method='PATCH',
            url=url,
            params=params,
            headers=headers,
            timeout=timeout,
            data=data,
            **kwargs,
        )



import enum


@enum.unique
class FiddlerTimestamp(str, enum.Enum):
    """Supported timestamp formats for events published to Fiddler"""

    EPOCH_MILLISECONDS = 'epoch milliseconds'
    EPOCH_SECONDS = 'epoch seconds'
    ISO_8601 = '%Y-%m-%d %H:%M:%S.%f'
    TIMEZONE = ISO_8601 + '%Z %z'
    INFER = 'infer'


@enum.unique
class FileType(str, enum.Enum):
    """Supported file types for ingestion"""

    CSV = '.csv'


@enum.unique
class ServerDeploymentMode(str, enum.Enum):
    F1 = 'f1'
    F2 = 'f2'


@enum.unique
class UploadType(str, enum.Enum):
    """To distinguish between dataset ingestion and event ingestion.
    Supposed to be only internally used.
    """

    DATASET = 'dataset'
    EVENT = 'event'



import argparse
import configparser
import json
import logging
import os
import logging

import requests

from fiddler._version import __version__
from fiddler.archive.build_container import build_cmd, register_cmd, run_cmd
from fiddler.archive.model_cmds import execute_cmd, explain_cmd
from fiddler.utils.exceptions import CommandOptionsError

LOG = logging.getLogger(__name__)

def require_arg(args, arg_name, reason=None):
    if getattr(args, arg_name) is None:

        raise CommandOptionsError(
            'option \'{}\' is required.{}'.format(
                arg_name.replace('_', '-'), f' {reason}.' if reason else ''
            )
        )
    return getattr(args, arg_name)


# Configuration schema:
# --------------------
# [default]
# organization = my_org
# auth_key = user_bearer_key_for_API
CONFIG_FILE = '{}/.fidl/config'.format(os.environ['HOME'])


def read_config():
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE)
    if 'default' not in config.sections():
        config['default'] = {}
    return config


def write_config(config):
    os.makedirs(os.path.dirname(CONFIG_FILE), exist_ok=True)
    with open(CONFIG_FILE, 'w') as out:
        config.write(out)


def api_url(args, api_method):
    """ Returns 'endpoint/api_method/org/project/model' """
    require_arg(args, 'project')
    return '%s/%s/%s/%s/%s' % (
        args.endpoint,
        api_method,
        args.org,
        args.project,
        args.model,
    )


def print_response(resp):
    for line in resp.iter_lines():
        # filter out keep-alive new lines
        if line:
            decoded_line = line.decode('utf-8')
            LOG.info(decoded_line)


def process_response(resp):
    """
    If the response is not successful, throws an
    exception with details about the error.
    Logs the response if the succeeds.
    """
    resp_str = resp.text
    try:
        resp_str = json.dumps(json.loads(resp_str), indent=4)
    except ValueError:
        # Response is not a json we print the raw text.
        pass
    if resp.status_code == requests.codes.ok:
        logging.info('Response from the service : %s', resp_str)
    else:
        logging.error('Request failed : %s', resp_str)
        resp.raise_for_status()


# Implementation of commands


def configure_cmd(args):
    config = read_config()
    section = config['default']
    for prop in ['organization', 'auth_key', 'endpoint']:
        cur = str(section.get(prop, ''))
        new = input(f'{prop}[{cur}] = ')
        if new:
            if prop == 'organization':
                section[prop] = new.lower()
            if prop == 'auth_key':
                section[prop] = new
            if prop == 'endpoint':
                section['endpoint'] = new
    write_config(config)


def parse_args(argv=None):
    """
    Parses command line arguments and runs the sub command.
    """

    common_args = argparse.ArgumentParser(description='Common Args', add_help=False)

    # Common args:
    # TODO: default project and other credentials
    #  could be stored in something like ~/.fiddler

    common_args.add_argument('--endpoint', help=argparse.SUPPRESS)

    common_args.add_argument(
        '--org', help='Organization name. ' 'Default is read from config.'
    )

    common_args.add_argument(
        '--auth-key', help='Bearer key used to ' 'authenticate REST API requests.'
    )

    model_args = argparse.ArgumentParser(
        description='Parser for --project & --model', add_help=False
    )

    model_args.add_argument(
        '--project', required=True, help='Project name in the organization. '
    )

    model_args.add_argument('--model', required=True, help='Name of the model')

    model_args.add_argument(
        '--port', required=False, default=5100, help='Port number to use'
    )

    build_args = argparse.ArgumentParser(
        description='Parser for --source', add_help=False
    )

    build_args.add_argument('--source', required=True, help='Name of the source dir')

    build_args.add_argument(
        '--train', required=False, default=False, help='Build container for training'
    )

    parser = argparse.ArgumentParser(description='Fiddler Archive (FAR)')

    parser.add_argument(
        '--version', action='version', version=f'%(prog)s {__version__}'
    )

    subparsers = parser.add_subparsers(
        title='Available commands', dest='command', metavar=''
    )

    # deploy command
    build_desc = 'build Fiddler Archive for the specified model'
    build_parser = subparsers.add_parser(
        'build',
        parents=[common_args, model_args, build_args],
        description=build_desc,
        help=build_desc,
    )
    build_parser.set_defaults(func=build_cmd)

    # run command

    run_desc = 'runs a Fiddler Archive'
    run_parser = subparsers.add_parser(
        'run',
        parents=[common_args, model_args],
        description=run_desc,
        help=run_desc,
    )
    run_parser.set_defaults(func=run_cmd)

    # execute command
    execute_args = argparse.ArgumentParser(
        description='Parser for --index', add_help=False
    )

    execute_args.add_argument(
        '--index', required=True, help='Index of the row to execute'
    )

    execute_args.add_argument(
        '--explanations', required=False, help='Type of explanations'
    )

    execute_desc = 'execute model wrapped in Fiddler Archive'
    execute_parser = subparsers.add_parser(
        'execute',
        parents=[common_args, model_args, execute_args],
        description=execute_desc,
        help=execute_desc,
    )
    execute_parser.set_defaults(func=execute_cmd)

    # explain command
    explain_desc = 'explain the specified row'
    explain_parser = subparsers.add_parser(
        'explain',
        parents=[common_args, model_args, execute_args],
        description=explain_desc,
        help=explain_desc,
    )
    explain_parser.set_defaults(func=explain_cmd)

    # configure command
    configure_desc = 'Update Fiddler configuration'
    configure_parser = subparsers.add_parser(
        'configure', description=configure_desc, help=configure_desc
    )
    configure_parser.set_defaults(func=configure_cmd)

    # register command

    register_args = argparse.ArgumentParser(
        description='container name', add_help=False
    )

    register_args.add_argument('--image', required=True, help='Container image')

    register_args.add_argument('--dataset', required=False, help='Name of the dataset')

    register_args.add_argument('--source', required=True, help='Name of the source dir')

    register_desc = 'Register a model on fiddler'
    register_parser = subparsers.add_parser(
        'register',
        parents=[common_args, model_args, register_args],
        description=register_desc,
        help=register_desc,
    )

    register_parser.set_defaults(func=register_cmd)

    args = parser.parse_args(argv)

    if args.command:
        try:
            # Read values configuration file for options
            # if they are not specified on cmdline.
            config = read_config()['default']
            if not args.org:
                args.org = config.get('organization', '')
                if not args.org:
                    raise CommandOptionsError(
                        '--organization option is required. '
                        'Run \'configure\' to set default organization'
                    )
            if not args.auth_key:
                args.auth_key = config.get('auth_key', '')

            if hasattr(args, 'data') and args.data:
                if args.data.startswith('@'):
                    with open(args.data[1:], 'r') as file:
                        args.data = file.read()

            if hasattr(args, 'endpoint') and args.endpoint:
                args.endpoint = args.endpoint.rstrip('/')
            else:
                args.endpoint = config.get('endpoint')

            ret = args.func(args)
            if ret is None or type(ret) != int:
                return 0
            else:
                return ret

        except CommandOptionsError as e:
            logging.error(f'{str(e)}')
            logging.error(
                f'Try `{parser.prog} {args.command} --help` for ' 'more help on usage'
            )
            return 1
        except Exception as e:
            logging.error(
                f'Command "{args.command}" failed '
                f'with the following error: {str(e)}',
                exc_info=True,
            )
            return 1
    else:
        parser.print_help()
        return 0


def main(argv=None):
    logging.basicConfig(
        level=logging.INFO, format='%(asctime)s %(levelname)-7s: %(message)s'
    )
    return parse_args(argv)


if __name__ == '__main__':
    exit(main())
